import requests
import socket
import logging


def parse(filename):

	config = {}

	with open(filename, 'r') as f:

		for line in f:
			if line[:1] == '#':
				continue
			(key, value) = line.split('=')
			config[key.strip()] = value.strip()

	return config

def instance_config(config):

	# Update with local instance data
	for update in (ec2_config, host_config):
		config = update(config)

	return config

def ec2_config(config):
	try:
		# Raises exception if non-existent, means we're not in EC2
		socket.gethostbyname('instance-data.ec2.internal')
		resp = requests.get('http://instance-data.ec2.internal/latest/user-data')
		try:
			json = resp.json()
			# Always override if user-data exists
			if u'identity' in json:
				config['identity'] = json[u'identity']
			if u'repository' in json:
				config['repository'] = json[u'repository']
		except Exception as e:
			logging.getLogger(__name__).error('Could not parse user-data: %s' % e)
	except Exception as e:
		pass
	return config

def host_config(config):
	# Fallback, only set if no ID is defined yet
	if 'identity' not in config:
		try:
			host = socket.getfqdn().split('.')
			host.reverse()
			config['identity'] = '.'.join(host)
		except:
			pass
	return config



import boto
import os
import sh
import shutil
import logging
import requests
import hashlib
from urlparse import urlparse
from repodeploy import config


def repository(url, work_dir, *args):

	repositories = {
		's3': S3Repository,
		'http': HttpRepository,
		'https': HttpRepository
	}

	try:
		parsed = urlparse(url)
		if parsed.scheme in repositories:
			return repositories[parsed.scheme](url, work_dir, *args)
		elif parsed.path.find('.git') > -1:
			return GitRepository(url, work_dir, *args)
	except Exception as e:
		raise Exception('Unable to create repository: %s (%s)' % (url, e))

	return None


class Repository(object):
	
	def __init__(self, url, work_dir, cfg):
		self.url = url
		self.work_dir = work_dir
		self.config = cfg
		self.log = logging.getLogger(__name__)
		self.link = False

	def current(self, path):
		"""
		Check the current configuration version from the repository to see if we need to update
		"""
		raise NotImplemented()

	def fetch(self, path):
		"""
		Fetch the resource at the specified path and return the local file or directory
		representing it in the working directory.
		"""
		raise NotImplemented()

	def workdir(self, name, remove=False):
		d = '%s/%s' % (self.work_dir, name)
		if os.path.exists(d) and remove:
			shutil.rmtree(d)
		if not os.path.exists(d):
			os.makedirs(d)
		return d
	
	def identity_path(self, identity):
		return '/'.join(identity.split('.'))

class S3Repository(Repository):

	def __init__(self, url, work_dir, cfg):

		super(S3Repository, self).__init__(url, work_dir, cfg)

		parsed = urlparse(url)
		self.bucket = parsed.hostname
		self.prefix = parsed.path

		if 'aws-access-key' in cfg:
			self.s3 = boto.connect_s3(cfg['aws-access-key'], cfg['aws-secret-key'])
		elif os.path.exists(os.path.expanduser('~/.amazon/account-key')):
			credentials = config.parse(os.path.expanduser('~/.amazon/account-key'))
			self.s3 = boto.connect_s3(credentials['accessKey'], credentials['secretKey'])
		else:
			self.s3 = boto.connect_s3()

	def current(self, identity):

		path = identity
		if self.prefix:
			path = '%s/%s' % (self.prefix, path)
		key = self.key(path)

		if key is not None:
			return key.etag.strip('"')

		return None

	def fetch(self, identity):

		cache = self.workdir('cache', True)
		path = identity
		if self.prefix:
			path = '%s/%s' % (self.prefix, path)
		key = self.key(path)

		if key is not None:

			outfile = '%s/%s' % (cache, os.path.basename(key.name))

			with open(outfile, 'w') as f:
				key.get_file(f)

			unpack_dir = self.workdir('unpacked', True)
			sh.unzip(outfile, _cwd=unpack_dir)
			return (key.etag.strip('"'), unpack_dir)

			raise Exception("Unable to unpack archive format '%s'" % ext)

		return (None, cache)

	def key(self, path):

		bucket = self.s3.get_bucket(self.bucket, validate=False)

		try:
			return bucket.get_key(path + '.zip')
		except:
			return None

class HttpRepository(Repository):

	def __init__(self, url, work_dir, cfg):
		super(HttpRepository, self).__init__(url, work_dir, cfg)

	def current(self, identity):
		return self.key(requests.head(self.make_url(identity)))

	def make_url(self, identity):
		return '%s/%s.zip' % (self.url, identity)

	def key(self, response):
		if 'etag' in response.headers:
			return response.headers['etag'].strip('"')
		else:
			m = hashlib.md5()
			m.update(response.headers['content-length'])
			m.update('/')
			m.update(response.headers['last-modified'])
			return m.hexdigest()


	def fetch(self, identity):

		cache = self.workdir('cache', True)
		r = requests.get(self.make_url(identity), stream=True)

		if r.status_code == 200:

			outfile = '%s/%s.zip' % (cache, identity)

			with open(outfile, 'w') as f:
				for chunk in r.iter_content(8192):
					f.write(chunk)

			unpack_dir = self.workdir('unpacked', True)
			sh.unzip(outfile, _cwd=unpack_dir)
			return (self.key(r), unpack_dir)

			raise Exception("Unable to unpack archive format '%s'" % ext)

		return (None, cache)

class GitRepository(Repository):

	def __init__(self, url, work_dir, cfg):
		super(GitRepository, self).__init__(url, work_dir, cfg)

		# Should use symlinks instead of moving since directory location isn't change
		self.link = True

		self.remote = url
		self.prefix = ''

		if not url.endswith('.git'):
			pos = url.rfind('.git') + 4
			self.remote = url[:pos]
			self.prefix = url[pos:]

		self.staging = self.workdir('git-work')
		self.local = self.workdir('git')

	def pull(self, directory):

		# Check for repository location change
		if os.path.exists('%s/%s' % (directory, '.git')):
			# Oh this will probably explode horribly at some point
			remote = str(sh.git('remote', '-v')).split('\n')[0].split('\t')[1].split(' ')[0]
			if remote != self.remote:
				# Location changed, wipe out and start again
				shutil.rmtree(directory)

		if not os.path.exists('%s/%s' % (directory, '.git')):
			# Fresh clone
			out = sh.git('clone', self.remote, directory, _cwd=os.path.dirname(directory))
			if out.exit_code != 0:
				raise Exception(str(out))
		else:
			# Ensure we revert to last good state (in case of moved files)
			out = sh.git('checkout', '.', _cwd=directory)
			if out.exit_code != 0:
				raise Exception(str(out))
			# Update from remote
			out = sh.git('pull', _cwd=directory)
			if out.exit_code != 0:
				raise Exception(str(out))
		# Parse latest revision from git
		return str(sh.git('rev-parse', 'HEAD', _cwd=directory)).strip()

	def current(self, identity):

		version = self.pull(self.staging)
		
		paths = [
			'%s%s/%s' % (self.staging, self.prefix, identity),
			'%s%s/%s' % (self.staging, self.prefix, self.identity_path(identity))
		]

		for path in paths:
			if os.path.exists(path):
				return version

		return None

	def fetch(self, identity):

		version = self.pull(self.local)

		paths = [
			'%s%s/%s' % (self.local, self.prefix, identity),
			'%s%s/%s' % (self.local, self.prefix, self.identity_path(identity))
		]

		for path in paths:
			if os.path.exists(path):
				return (version, path)

		return (None, None)



import sys
import os
import time
import datetime
import sched
import sh
import shutil
import logging
import subprocess
from croniter import croniter
from repodeploy import repo


class Deployer:

	def __init__(self, config):

		self.log = logging.getLogger(__name__)

		self.config = config
		self.version = None

		self.identity = config['identity'] if 'identity' in config else None
		self.directory = os.path.abspath(config['directory']) if 'directory' in config else os.getcwd()
		self.action = config['action'] if 'action' in config else None

		self.versionfile = '%s/current.version' % self.directory
		self.currentdir = '%s/current' % self.directory
		self.workdir = '%s/work' % self.directory

		# Create work dir
		if not os.path.exists(self.workdir):
			os.makedirs(self.workdir)

		# Create current config if nonexistent
		if not os.path.lexists(self.currentdir):
			os.makedirs(self.currentdir)
		elif os.path.exists(self.versionfile):
			with open(self.versionfile, 'r') as f:
				version = f.readline().strip()
				if version is not None:
					self.version = version

		try:
			self.repository = repo.repository(config['repository'], self.workdir, self.config)
		except Exception as e:
			self.log.error('unable to initialize configuration repository: %s (%s)' % (config['repository'], e))
			sys.exit(1)

		if self.identity is None:
			self.log.error('no identity set, exiting')
			sys.exit(1)

		self.log.info('identity=%s' % self.identity)
		self.log.info('repository=%s' % self.repository.url)

	def run(self):

		scheduler = sched.scheduler(time.time, time.sleep)
		cron = croniter(self.config['schedule']) if 'schedule' in self.config else croniter('* * * * *')

		self.check_config()

		while True:
			try:
				scheduler.enterabs(cron.get_next(float), 1, self.check_config, ())
				scheduler.run()
			except KeyboardInterrupt as k:
				break
			except Exception as e:
				self.log.error("exception during update: %s" % e)
	
	def check_config(self):

		self.log.debug('checking for config updates')
		version = self.repository.current(self.identity)
		if version is not None:
			if version != self.version:
				self.log.info('updating to version %s' % version)
				self.update_config()
				self.log.info('update complete')
			else:
				self.log.debug('on latest version %s' % version)
			return True
		else:
			self.log.warn('no configuration found for identity: %s' % self.identity)
			return False

	def update_config(self):

		try:

			(version, directory) = self.repository.fetch(self.identity)

			if not directory:
				self.log.warn('update failed to properly unpack, skipping')
				return False

			# Pre-update scripts
			if not self.run_hooks(self.config['pre_hooks'], current=directory, previous=self.currentdir):
				self.log.warn('pre-update hooks failed, update blocked')
				return False

			save = None

			# Repository is link-safe, target directory does not change location
			if self.repository.link:
				# Repository probably changed, full reset
				if not os.path.islink(self.currentdir):
					shutil.rmtree(self.currentdir)
				else:
					save = os.path.realpath(self.currentdir)
				if save == directory:
					save = None
				else:
					sh.ln('-sfT', directory, self.currentdir)

			# Unpacked directories may move, need to copy instead of link
			else:
				save = '%s/config.save' % self.workdir
				# Repository probably changed, full reset
				if os.path.islink(self.currentdir):
					os.unlink(self.currentdir)
				elif os.path.exists(self.currentdir):
					self.move_contents(self.currentdir, save)
				self.move_contents(directory, self.currentdir)
				shutil.rmtree(directory)

			self.log.debug('activated configuration version: %s' % version)

			# Post-update scripts
			if not self.run_hooks(self.config['post_hooks'], current=self.currentdir, previous=save):
				if save and os.path.exists(save):
					self.log.warn('post-update hooks failed, reverting to previous configuration')
					if self.repository.link:
						target = os.path.realpath(self.currentdir)
						sh.ln('-sf', save, self.currentdir)
						# Should always be true at this point
						if target != save:
							shutil.rmtree(target)
					else:
						self.move_contents(save, self.currentdir)
						shutil.rmtree(save)
					# Re-run post-update hooks after revert
					if not self.run_hooks(self.config['post_hooks'], current=self.currentdir):
						self.log.error('post-update hooks failed after revert, application may be unstable')
				else:
					self.log.error('post-update hooks failed, application may be unstable')
				return False
			else:
				if save is not None and os.path.exists(save):
					shutil.rmtree(save)
				self.version = version
				with open(self.versionfile, 'w') as f:
					f.write(self.version)
				return True

		except Exception as e:
			self.log.error('could not update configuration: %s' % unicode(e))

		return False

	def move_contents(self, src, dest):
		"""
		Some apps resolve the final link when monitoring directories, so we
		can't just change symlinks, we need to move files.
		"""

		if not os.path.exists(dest):
			os.makedirs(dest)
		else:
			# Remove files in dest (if exists)
			destfiles = os.listdir(dest)
			if len(destfiles):
				sh.rm('-rf', *destfiles, _cwd=dest)

		# Move src files to dest
		srcfiles = os.listdir(src)
		if len(srcfiles):
			srcfiles.append(dest)
			sh.mv(*srcfiles, _cwd=src)

		return True

	def run_hooks(self, hook_dir, current=None, previous=None):
		for script in os.listdir(hook_dir):
			full = '%s/%s' % (hook_dir, script)
			if os.access(full, os.X_OK):
				self.log.info('running update hook: %s' % full)
				# Pass in current/previous config directories
				env = os.environ.copy()
				env['CURRENT_CONFIG'] = current if current else ''
				env['PREVIOUS_CONFIG'] = previous if previous else ''
				p = subprocess.Popen(full, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)
				(out, err) = p.communicate()
				if p.returncode > 0:
					for line in out.split('\n'):
						if line:
							self.log.warn(line)
					return False
		return True
